@@ -40,6 +40,7 @@ SKIP_HOOKS_MD_KEY = 'x-heptapod-skip-gl-hooks'
 ACCEPT_MR_IID_KEY = 'x-heptapod-accept-mr-iid'
 # TODO expose it in heptpod.wsgi and import it here
 PY_HEPTAPOD_SKIP_HOOKS = b'HEPTAPOD_SKIP_ALL_GITLAB_HOOKS'
+HEPTAPOD_PERMISSION_KEY = 'x-heptapod-perm'
 
 logger = logging.getLogger(__name__)
 
@@ -178,6 +179,8 @@ class HGitalyServicer:
         """
         mirroring = native = skip_gl_hooks = False
         accept_mr_iid = None
+        self.heptapod_permission = None
+
         for md in context.invocation_metadata():
             if md.key == HG_GIT_MIRRORING_MD_KEY:
                 mirroring = md.value.strip().lower() == 'true'
@@ -187,6 +190,8 @@ class HGitalyServicer:
                 skip_gl_hooks = md.value.strip().lower() == 'true'
             elif md.key == ACCEPT_MR_IID_KEY:
                 accept_mr_iid = md.value.strip()
+            elif md.key == HEPTAPOD_PERMISSION_KEY:
+                self.heptapod_permission = md.value.strip()
 
         mirroring = mirroring or not native
 
