@@ -221,12 +221,20 @@ class BaseRpcHelper:
 
     Will handle comparison between Gitaly and (HGitaly or RHGitaly) as well
     as comparison beween HGitaly and RHGitaly, or whatever comes next.
+
+    :param streaming requests_with_header: yield a first request
+      controlled by `first_request_sub_message`, and then as iterable of
+      `dicts`
+    :param streaming_request_field: produce a stream of requests by
+       cutting a single incoming field into chunks.
     """
 
     def __init__(self, comparison, stub_cls, method_name, request_cls,
                  repository_arg=True,
                  request_defaults=None,
                  feature_flags=(),
+                 streaming_requests_raw=False,
+                 streaming_requests_with_header=None,
                  streaming_request_field=None,
                  streaming_request_chunk_size=1,
                  streaming=False):
@@ -236,6 +244,8 @@ class BaseRpcHelper:
         self.request_cls = request_cls
         self.streaming = streaming
         self.streaming_request_field = streaming_request_field
+        self.streaming_requests_raw = streaming_requests_raw
+        self.streaming_requests_with_header = streaming_requests_with_header
         self.streaming_request_chunk_size = streaming_request_chunk_size
         self.repository_arg = repository_arg
 
@@ -272,6 +282,14 @@ class BaseRpcHelper:
             # case with `UpdateRemoteMirror`
             yield self.request_cls(**req)
 
+    def stream_requests_with_header(self, tail_requests=(), **kwargs):
+        """Tailored for UserCommtiFiles. Could be useful elsewhere."""
+        header_attr, header_cls = self.streaming_requests_with_header
+        yield self.request_cls(**{header_attr: header_cls(**kwargs)})
+
+        for req in tail_requests:
+            yield self.request_cls(**req)
+
     def rpc(self, backend, **kwargs):
         if self.repository_arg:
             if backend == 'hg':
@@ -281,6 +299,10 @@ class BaseRpcHelper:
 
         if self.streaming_request_field is not None:
             request = self.stream_requests(**kwargs)
+        elif self.streaming_requests_raw:
+            request = iter(kwargs['requests'])
+        elif self.streaming_requests_with_header is not None:
+            request = self.stream_requests_with_header(**kwargs)
         else:
             request = self.request_cls(**kwargs)
 
@@ -710,10 +732,10 @@ class RpcHelper(BaseRpcHelper):
         self.apply_request_defaults(hg_kwargs)
 
         git_kwargs = self.request_kwargs_to_git(hg_kwargs)
-        with pytest.raises(grpc.RpcError) as exc_info_hg:
-            self.rpc('hg', **hg_kwargs)
         with pytest.raises(grpc.RpcError) as exc_info_git:
             self.rpc('git', **git_kwargs)
+        with pytest.raises(grpc.RpcError) as exc_info_hg:
+            self.rpc('hg', **hg_kwargs)
         self.assert_compare_grpc_exceptions(
             exc_info_hg.value,
             exc_info_git.value,
