@@ -7,26 +7,60 @@
 import pytest
 import grpc
 import re
-from hgitaly.git import EMPTY_TREE_OID
+
+from mercurial_testhelpers import (
+    as_bytes,
+)
+
+from hgitaly.git import (
+    EMPTY_TREE_OID,
+    NULL_BLOB_OID,
+)
+from hgitaly.oid import tree_oid
+from hgitaly.revision import (
+    gitlab_revision_hash,
+)
+from hgitaly.stub.commit_pb2 import (
+    TreeEntryRequest,
+    TreeEntryResponse,
+)
 from hgitaly.stub.diff_pb2 import (
     CommitDeltaRequest,
     CommitDiffRequest,
     DiffStatsRequest,
     FindChangedPathsRequest,
+    GetPatchIDRequest,
     RawDiffRequest,
     RawPatchRequest,
 )
+from hgitaly.stub.commit_pb2_grpc import CommitServiceStub
 from hgitaly.stub.diff_pb2_grpc import DiffServiceStub
+from .test_blob_tree import oid_mapping  # TODO make it a fixture method
 
 from . import skip_comparison_tests
 if skip_comparison_tests():  # pragma no cover
     pytestmark = pytest.mark.skip
 
+parametrize = pytest.mark.parametrize
+
 INDEX_LINE_REGEXP = re.compile(br'^index \w+\.\.\w+( \d+)?$')
 
 
-def test_compare_raw_diff(gitaly_comparison):
-    fixture = gitaly_comparison
+def generate_diff_stubs(fixture, hg_server='hgitaly'):
+    if hg_server == 'hgitaly':
+        hgitaly_channel = fixture.hgitaly_channel
+    elif hg_server == 'rhgitaly':
+        hgitaly_channel = fixture.rhgitaly_channel
+
+    return dict(
+        git=DiffServiceStub(fixture.gitaly_channel),
+        hg=DiffServiceStub(hgitaly_channel)
+    )
+
+
+@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
+def test_compare_raw_diff(gitaly_rhgitaly_comparison, hg_server):
+    fixture = gitaly_rhgitaly_comparison
     wrapper_repo = fixture.gitaly_repo
     git_repo = fixture.git_repo
     wrapper = fixture.hg_repo_wrapper
@@ -59,10 +93,7 @@ def test_compare_raw_diff(gitaly_comparison):
     # o  0 Some bar
     #
 
-    diff_stubs = dict(
-        git=DiffServiceStub(fixture.gitaly_channel),
-        hg=DiffServiceStub(fixture.hgitaly_channel)
-    )
+    diff_stubs = generate_diff_stubs(fixture, hg_server)
 
     def do_rpc(vcs, left_cid, right_cid):
         if vcs == 'git':
@@ -110,10 +141,10 @@ def test_compare_raw_diff(gitaly_comparison):
     sha_not_exists = b'deadnode' * 5
     with pytest.raises(grpc.RpcError) as exc_info:
         do_rpc('hg', sha_not_exists, ctx2.hex())
-    assert exc_info.value.code() == grpc.StatusCode.UNKNOWN
+    assert exc_info.value.code() == grpc.StatusCode.INTERNAL
     with pytest.raises(grpc.RpcError) as exc_info:
         do_rpc('git', sha_not_exists, ctx2.hex())
-    assert exc_info.value.code() == grpc.StatusCode.UNKNOWN
+    assert exc_info.value.code() == grpc.StatusCode.INTERNAL
 
     # case 5: EMPTY_TREE_OID to represent null commit on the left
     hg_resp_lines = do_rpc('hg', EMPTY_TREE_OID, ctx0.hex()).split(b'\n')
@@ -138,8 +169,9 @@ def test_compare_raw_diff(gitaly_comparison):
     assert hg_resp_lines == git_resp_lines
 
 
-def test_compare_raw_patch(gitaly_comparison):
-    fixture = gitaly_comparison
+@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
+def test_compare_raw_patch(gitaly_rhgitaly_comparison, hg_server):
+    fixture = gitaly_rhgitaly_comparison
     gitaly_repo = fixture.gitaly_repo
     git_repo = fixture.git_repo
     wrapper = fixture.hg_repo_wrapper
@@ -149,10 +181,7 @@ def test_compare_raw_patch(gitaly_comparison):
                                message="Some bar")
     git_sha0 = git_repo.branches()[gl_branch]['sha']
 
-    diff_stubs = dict(
-        git=DiffServiceStub(fixture.gitaly_channel),
-        hg=DiffServiceStub(fixture.hgitaly_channel)
-    )
+    diff_stubs = generate_diff_stubs(fixture, hg_server)
 
     def do_rpc(vcs, left_cid, right_cid):
         request = RawPatchRequest(
@@ -189,11 +218,12 @@ def test_compare_commit_diff(gitaly_comparison):
     wrapper.command(b'mv', wrapper.repo.root + b'/bar',
                     wrapper.repo.root + b'/zar')
     wrapper.command(b'ci', message=b'Rename bar to zar')
-    ctx3 = wrapper.commit_file('zoo', content='I am in \nzoo\n',
+    ctx3 = wrapper.commit_file('zoo', content='I am in \n zoo\n',
                                message='Added zoo')
     git_sha3 = git_repo.branches()[gl_branch]['sha']
-    wrapper.commit_file('zoo', content='I am in\nzoo\n',
-                        message='zoo: remove whitespace')
+    ctx4 = wrapper.commit_file('zoo', content='I am in\nzoo\n',
+                               message='zoo: change whitespace')
+    git_sha4 = git_repo.branches()[gl_branch]['sha']
     ctx5 = wrapper.commit_file('too', content='I am in\ntoo\n',
                                message='Added too')
     git_sha5 = git_repo.branches()[gl_branch]['sha']
@@ -234,6 +264,11 @@ def test_compare_commit_diff(gitaly_comparison):
             final.append(resp)
         return final
 
+    def rpc_diff_size(*args, **kwargs):
+        """Perform RPC and return the total length of patch data it gives."""
+        return sum(len(chunk.raw_patch_data)
+                   for chunk in do_rpc(*args, **kwargs))
+
     # case 1: actual test (rename + new file + content change)
     hg_resp = do_rpc('hg', left_cid=ctx1.hex(),
                      right_cid=ctx3.hex())
@@ -310,6 +345,33 @@ def test_compare_commit_diff(gitaly_comparison):
     assert hg_resp == git_resp
     assert len(hg_resp) == 2
 
+    # case 6: whitespace options
+    WHITESPACE_IGNORE = CommitDiffRequest.WHITESPACE_CHANGES_IGNORE
+    WHITESPACE_IGNORE_ALL = CommitDiffRequest.WHITESPACE_CHANGES_IGNORE_ALL
+    hg_resp = do_rpc('hg', left_cid=ctx3.hex(),
+                     right_cid=ctx4.hex(),
+                     whitespace_changes=WHITESPACE_IGNORE)
+    git_resp = do_rpc('git', left_cid=git_sha3,
+                      right_cid=git_sha4,
+                      whitespace_changes=WHITESPACE_IGNORE)
+    # Git and Mercurial differ in how they represent the line with
+    # ignored changes. Both mark it as unchanged in the patch, but
+    # Git displays the new value for context Mercurial displays the old value:
+    hg_raw = hg_resp[0].raw_patch_data.splitlines()
+    # we don't change the first char (' ', '-', or '+')
+    hg_raw[1] = hg_raw[1][:-1]
+    hg_raw.append(b'')  # for final EOL
+    hg_resp[0].raw_patch_data = b'\n'.join(hg_raw)
+    assert hg_resp == git_resp
+
+    hg_size = rpc_diff_size('hg', left_cid=ctx3.hex(),
+                            right_cid=ctx4.hex(),
+                            whitespace_changes=WHITESPACE_IGNORE_ALL)
+    git_size = rpc_diff_size('git', left_cid=git_sha3,
+                             right_cid=git_sha4,
+                             whitespace_changes=WHITESPACE_IGNORE_ALL)
+    assert hg_size == git_size
+
 
 def test_compare_commit_delta(gitaly_comparison):
     fixture = gitaly_comparison
@@ -467,17 +529,19 @@ def test_compare_find_changed_paths(gitaly_comparison):
     wrapper = fixture.hg_repo_wrapper
 
     gl_branch = b'branch/default'
-    (wrapper.path / 'foo').write_text('foo content')
+    (wrapper.path / 'sub').mkdir()
+    (wrapper.path / 'sub/foo').write_text('foo content')
     (wrapper.path / 'bar').write_text('bar content')
-    ctx0 = wrapper.commit(rel_paths=['foo', 'bar'],
+    ctx0 = wrapper.commit(rel_paths=['sub/foo', 'bar'],
                           add_remove=True)
     git_sha0 = git_repo.branches()[gl_branch]['sha']
 
     (wrapper.path / 'zoo').write_text('zoo content')
-    (wrapper.path / 'foo').write_text('foo content modified')
+    (wrapper.path / 'zoo').chmod(0o755)
+    (wrapper.path / 'sub/foo').write_text('foo content modified')
     (wrapper.path / 'bar').unlink()
     wrapper.command(b'rm', wrapper.repo.root + b'/bar')
-    ctx1 = wrapper.commit(rel_paths=['foo', 'bar', 'zoo'],
+    ctx1 = wrapper.commit(rel_paths=['sub/foo', 'bar', 'zoo'],
                           add_remove=True)
     git_sha1 = git_repo.branches()[gl_branch]['sha']
 
@@ -486,20 +550,104 @@ def test_compare_find_changed_paths(gitaly_comparison):
         hg=DiffServiceStub(fixture.hgitaly_channel)
     )
 
-    def do_rpc(vcs, commits):
-        request = FindChangedPathsRequest(
-            repository=wrapper_repo,
-            commits=commits,
-        )
+    oid_to_git = oid_mapping(fixture,
+                             ((ctx.hex(), path, 'blob')
+                              for ctx, path in (
+                                      (ctx0, b'sub/foo'),
+                                      (ctx1, b'sub/foo'),
+                                      (ctx0, b'bar'),
+                                      (ctx1, b'zoo'),
+                              )))
+    oid_to_git[NULL_BLOB_OID] = NULL_BLOB_OID
+
+    def do_rpc_inner(vcs, **req_kwargs):
+        """Transitional to accomodate deprecated and new style"""
+
+        request = FindChangedPathsRequest(repository=wrapper_repo,
+                                          **req_kwargs)
         response = diff_stubs[vcs].FindChangedPaths(request)
         final = []
         for resp in response:
             paths = sorted(resp.paths, key=lambda o: o.path)
             final.append(paths)
+            if vcs == 'hg':
+                for path in paths:
+                    if path.old_blob_id:
+                        path.old_blob_id = oid_to_git[path.old_blob_id]
+                    if path.new_blob_id:
+                        path.new_blob_id = oid_to_git[path.new_blob_id]
         return final
 
+    def do_rpc_depr(vcs, commits):
+        return do_rpc_inner(vcs, commits=commits)
+
+    CommitRequest = FindChangedPathsRequest.Request.CommitRequest
+
+    def do_rpc_commits(vcs, commits, compare_to=(), **kw):
+        """Using the new (as of Gitaly 15.2 CommitRequest message.
+
+        for more cases, use do_rpc_inner directly
+        :param compare_to: if given, will be used in all requests.
+        """
+        return do_rpc_inner(
+            vcs,
+            requests=[
+                FindChangedPathsRequest.Request(
+                    commit_request=CommitRequest(
+                        commit_revision=c,
+                        parent_commit_revisions=compare_to))
+                for c in commits],
+            **kw
+        )
+
+    TreeRequest = FindChangedPathsRequest.Request.TreeRequest
+
+    def git_tree(revision, path):
+        stub = CommitServiceStub(fixture.gitaly_channel)
+        resps = list(stub.TreeEntry(
+            TreeEntryRequest(repository=wrapper_repo,
+                             revision=revision,
+                             path=path)
+        ))
+        assert len(resps) == 1
+        assert resps[0].type == TreeEntryResponse.ObjectType.TREE
+        return resps[0].oid
+
+    def hg_tree(revision, path):
+        repo = wrapper.repo
+        return tree_oid(repo,
+                        gitlab_revision_hash(repo, revision).decode('ascii'),
+                        path)
+
+    tree_finders = dict(hg=hg_tree, git=git_tree)
+
+    def do_rpc_tree(vcs, left_commit, right_commit, path):
+        # for now we support only the case where all trees are the same subdir
+        # in different commits.
+
+        tree_finder = tree_finders[vcs]
+        return do_rpc_inner(
+            vcs,
+            requests=[
+                FindChangedPathsRequest.Request(
+                    tree_request=TreeRequest(
+                        left_tree_revision=tree_finder(left_commit, path),
+                        right_tree_revision=tree_finder(right_commit, path))
+                )
+            ]
+        )
+
+    assert (do_rpc_tree('git', git_sha0, git_sha1, b'sub')
+            ==
+            do_rpc_tree('hg', ctx0.hex(), ctx1.hex(), b'sub')
+            )
+
     # case 1: actual test
-    assert do_rpc('git', [git_sha1]) == do_rpc('hg', [ctx1.hex()])
+    assert do_rpc_depr('git', [git_sha1]) == do_rpc_depr('hg', [ctx1.hex()])
+    assert (do_rpc_commits('git', [git_sha1])
+            ==
+            do_rpc_commits('hg', [ctx1.hex()])
+            )
 
     # Note: khanchi97: As per the command used by gitaly for FindChangedPaths
     # cmd: `git diff-tree --stdin -z -m -r --name-status --no-renames`
@@ -509,9 +657,150 @@ def test_compare_find_changed_paths(gitaly_comparison):
 
     # case 2: when commit_id does not correspond to a commit
     sha_not_exists = b'deadnode' * 5
+
     with pytest.raises(grpc.RpcError) as exc_info_hg:
-        do_rpc('hg', [ctx0.hex(), sha_not_exists])
+        do_rpc_depr('hg', [ctx0.hex(), sha_not_exists])
     with pytest.raises(grpc.RpcError) as exc_info_git:
-        do_rpc('git', [git_sha0, sha_not_exists])
+        do_rpc_depr('git', [git_sha0, sha_not_exists])
     assert exc_info_hg.value.code() == exc_info_git.value.code()
     assert exc_info_hg.value.details() == exc_info_git.value.details()
+
+    with pytest.raises(grpc.RpcError) as exc_info_hg:
+        do_rpc_commits('hg', [ctx0.hex(), sha_not_exists])
+    with pytest.raises(grpc.RpcError) as exc_info_git:
+        do_rpc_commits('git', [git_sha0, sha_not_exists])
+    assert exc_info_hg.value.code() == exc_info_git.value.code()
+    assert exc_info_hg.value.details() == exc_info_git.value.details()
+
+    # case 3: tree requests
+    assert (do_rpc_tree('git', git_sha0, git_sha1, b'sub')
+            ==
+            do_rpc_tree('hg', ctx0.hex(), ctx1.hex(), b'sub')
+            )
+
+    # can't compare with copies because Gitaly actually fails to
+    # detect them (it does not pass the `-C` flag to `git diff-tree`,
+    # at least as of gitaly@2b069d853)
+    # not sure if that counts as a bug on their side or is really
+    # intended, even though the copied case is part of protocol.
+
+    # case 4: several requests, checking our analysis that results are
+    # not merged.
+    # Let's begin with the extreme case, twice the same commit:
+    assert (do_rpc_commits('git', [git_sha1, git_sha1])
+            ==
+            do_rpc_commits('hg', [ctx1.hex(), ctx1.hex()])
+            )
+    # Less disturbing, let's make a new commit, modifying `foo` once again
+    (wrapper.path / 'sub/foo').write_text('foo content re modified')
+    hg_sha2 = wrapper.commit(rel_paths=['sub']).hex()
+    git_sha2 = git_repo.branches()[gl_branch]['sha']
+    fixture.invalidate()
+
+    oid_to_git.update(oid_mapping(fixture,
+                                  [(hg_sha2, b'sub/foo', 'blob')]))
+    assert (do_rpc_commits('git', [git_sha1, git_sha2])
+            ==
+            do_rpc_commits('hg', [ctx1.hex(), hg_sha2])
+            )
+
+    # case 4: merge
+    # In case of several parents, `git diff-tree -m`, on which the Gitaly
+    # implementation is based would display the diffs against each parent,
+    # one after the other.
+
+    other_ctx = wrapper.commit_file('new', branch='other', parent=ctx0,
+                                    content="new file in branch")
+    wrapper.update(2)
+    wrapper.command('merge', other_ctx.hex())
+    # the foo file will appeared to be modified when compared with each parent
+    (wrapper.path / 'sub/foo').write_text('something else entirely')
+    # TODO implement comiit with "always matcher" in mercurial-testhelpers
+    # (merge commits don't allow specifying paths, even if we list all)
+    wrapper.command('commit', message=b'the merge')
+    hg_sha3 = wrapper.repo[b'tip'].hex()
+    git_sha3 = git_repo.branches()[gl_branch]['sha']
+    fixture.invalidate()
+
+    oid_to_git.update(
+        oid_mapping(fixture,
+                    ((sha, path, 'blob') for sha, path in (
+                        (hg_sha3, b'new'),
+                        (hg_sha3, b'sub/foo'),
+                        (hg_sha3, b'zoo'),
+                        (other_ctx.hex(), b'bar'),
+                        (other_ctx.hex(), b'new'),
+                        (other_ctx.hex(), b'sub/foo'),
+                        ))))
+    assert (do_rpc_commits('git', [git_sha3])
+            ==
+            do_rpc_commits('hg', [hg_sha3])
+            )
+
+    # case 5: passing "parents" explicitely
+    hg_sha4 = wrapper.commit_file("toto").hex()
+    git_sha4 = git_repo.branches()[gl_branch]['sha']
+    fixture.invalidate()
+
+    oid_to_git.update(
+        oid_mapping(fixture,
+                    ((sha, path, 'blob') for sha, path in (
+                        (hg_sha4, b'new'),
+                        (hg_sha4, b'sub/foo'),
+                        (hg_sha4, b'toto'),
+                        (hg_sha4, b'zoo'),
+                        ))))
+    assert (do_rpc_commits('git', [git_sha4], compare_to=[git_sha0])
+            ==
+            do_rpc_commits('hg', [hg_sha4], compare_to=[ctx0.hex()])
+            )
+
+    # case 6: with renames
+    repo_path = wrapper.path
+    wrapper.command(
+        'mv', as_bytes(repo_path / 'zoo'), as_bytes(repo_path / 'zaz')
+    )
+    wrapper.command('commit', message=b'cp/mv')
+    hg_sha5 = wrapper.repo[b'tip'].hex()
+    git_sha5 = git_repo.branches()[gl_branch]['sha']
+    fixture.invalidate()
+    oid_to_git.update(oid_mapping(fixture,
+                                  [(hg_sha5, b'zaz', 'blob')]))
+
+    for find_renames in (False, True):
+        assert (do_rpc_commits('git', [git_sha5], find_renames=find_renames)
+                ==
+                do_rpc_commits('hg', [hg_sha5], find_renames=find_renames)
+                )
+
+
+def test_compare_get_patch_id(gitaly_comparison):
+    fixture = gitaly_comparison
+    wrapper = fixture.hg_repo_wrapper
+    rpc_helper = fixture.rpc_helper(
+        stub_cls=DiffServiceStub,
+        method_name='GetPatchID',
+        request_cls=GetPatchIDRequest,
+        request_sha_attrs=['old_revision', 'new_revision'],
+    )
+    assert_compare = rpc_helper.assert_compare
+    assert_compare_errors = rpc_helper.assert_compare_errors
+
+    hex0 = wrapper.commit_file('regular', content='foo\n').hex()
+    hex1 = wrapper.commit_file('regular', content='bar\n', topic='bar').hex()
+
+    gl_branch = b'branch/default'
+    gl_topic = b'topic/default/bar'
+
+    # regular operation with symbolic or sha revisions
+    assert_compare(old_revision=gl_branch, new_revision=gl_topic)
+    assert_compare(old_revision=hex0, new_revision=gl_topic)
+    assert_compare(old_revision=gl_branch, new_revision=hex1)
+
+    # errors on unknown revisions (for now INTERNAL with Gitaly, highly
+    # probable that it will become something else and that details really
+    # do not matter)
+    assert_compare_errors(old_revision=gl_branch, new_revision=b'unknown',
+                          same_details=False)
+    assert_compare_errors(old_revision=b'unknown', new_revision=gl_branch,
+                          same_details=False)
