@@ -382,7 +382,7 @@ def test_compare_commit_files(comparison):
         branch_name=b'branch/default',
         tail_requests=ucf_actions(dict(
             action=ActionType.CREATE,
-            file_path=b'foo',
+            file_path=b'/foo',
             content=b'foo content')),
     )
 
@@ -498,8 +498,7 @@ def test_compare_commit_files(comparison):
         ))
     )
 
-    # invalid path (Gitaly does not reject absolute paths and actually
-    # seems to ignore them, thankfully)
+    # invalid path
     for path in ('double//slash', '../../../good.joke'):
         for action in ('CREATE', 'CREATE_DIR', 'CHMOD',
                        'UPDATE', 'DELETE', 'MOVE'):
